@@ -6,11 +6,7 @@ import struct
 import sys
 
 __title__ = 'puzpy'
-__version__ = '0.2.1'
-__author__ = 'Alex DeJarnatt'
-__author_email__ = 'adejarnatt@gmail.com'
-__maintainer__ = 'Simeon Visser'
-__maintainer_email__ = 'simeonvisser@gmail.com'
+__version__ = '0.3.2'
 __license__ = 'MIT'
 __copyright__ = 'Copyright 2009 Alex DeJarnatt'
 
@@ -21,8 +17,8 @@ if PY3:
     str = str
     range = range
 else:
-    str = unicode
-    range = xrange
+    str = unicode  # noqa: F821
+    range = xrange  # noqa: F821
 
 HEADER_FORMAT = '''<
              H 11s        xH
@@ -37,10 +33,14 @@ EXTENSION_HEADER_FORMAT = '< 4s  H H '
 MASKSTRING = 'ICHEATED'
 
 ENCODING = 'ISO-8859-1'
+ENCODING_UTF8 = 'UTF-8'
+ENCODING_ERRORS = 'strict'  # raises an exception for bad chars; change to 'replace' for laxer handling
 
-ACROSSDOWN = 'ACROSS&DOWN'
+ACROSSDOWN = b'ACROSS&DOWN'
 
 BLACKSQUARE = '.'
+BLACKSQUARE2 = ':'
+BLANKSQUARE = '-'
 
 
 def enum(**enums):
@@ -57,6 +57,8 @@ PuzzleType = enum(
 SolutionState = enum(
     # solution is available in plaintext
     Unlocked=0x0000,
+    # solution is not present in the file
+    NotProvided=0x0002,
     # solution is locked (scrambled) with a key
     Locked=0x0004
 )
@@ -79,14 +81,19 @@ GridMarkup = enum(
 Extensions = enum(
     # grid of rebus indices: 0 for non-rebus;
     # i+1 for key i into RebusSolutions map
+    # should be same size as the grid
     Rebus=b'GRBS',
+
     # map of rebus solution entries eg 0:HEART;1:DIAMOND;17:CLUB;23:SPADE;
     RebusSolutions=b'RTBL',
-    # user's rebus entries
+
+    # user's rebus entries, same format as RebusSolutions
     RebusFill=b'RUSR',
+
     # timer state: 'a,b' where a is the number of seconds elapsed and
     # b is a boolean (0,1) for whether the timer is running
     Timer=b'LTIM',
+
     # grid cell markup: previously incorrect: 0x10;
     # currently incorrect: 0x20,
     # hinted: 0x40,
@@ -98,22 +105,39 @@ Extensions = enum(
 def read(filename):
     """
     Read a .puz file and return the Puzzle object.
-    throws PuzzleFormatError if there's any problem with the file format.
+    raises PuzzleFormatError if there's any problem with the file format.
     """
     with open(filename, 'rb') as f:
         return load(f.read())
 
 
+def read_text(filename):
+    """
+    Read an Across Lite .txt text format file and return the Puzzle object.
+    raises PuzzleFormatError if there's any problem with the file format.
+    """
+    with open(filename, 'r', encoding='utf-8', errors='replace') as f:
+        return load_text(f.read())
+
+
 def load(data):
     """
     Read .puz file data and return the Puzzle object.
-    throws PuzzleFormatError if there's any problem with the file format.
+    raises PuzzleFormatError if there's any problem with the file format.
     """
     puz = Puzzle()
     puz.load(data)
     return puz
 
 
+def load_text(text):
+    """
+    Parse Across Lite Text format from a string or bytes and return a Puzzle object.
+    raises PuzzleFormatError if there's any problem with the format.
+    """
+    return from_text_format(text)
+
+
 class PuzzleFormatError(Exception):
     """
     Indicates a format error in the .puz file. May be thrown due to
@@ -126,18 +150,21 @@ class PuzzleFormatError(Exception):
 class Puzzle:
     """Represents a puzzle
     """
-    def __init__(self):
+    def __init__(self, version='1.3'):
         """Initializes a blank puzzle
         """
-        self.preamble = ''
-        self.postscript = ''
+        self.preamble = b''
+        self.postscript = b''
         self.title = ''
         self.author = ''
         self.copyright = ''
         self.width = 0
         self.height = 0
-        self.version = b'1.3'
-        self.fileversion = b'1.3\0'  # default
+        if not isinstance(version, bytes):
+            version = bytes(version.encode('utf8'))
+        self.version = version
+        self.fileversion = version + b'\0'
+        self.encoding = ENCODING
         # these are bytes that might be unused
         self.unk1 = b'\0' * 2
         self.unk2 = b'\0' * 12
@@ -159,7 +186,7 @@ class Puzzle:
         # advance to start - files may contain some data before the
         # start of the puzzle use the ACROSS&DOWN magic string as a waypoint
         # save the preamble for round-tripping
-        if not s.seek_to(ACROSSDOWN.encode(ENCODING), -2):
+        if not s.seek_to(ACROSSDOWN, -2):
             raise PuzzleFormatError("Data does not appear to represent a "
                                     "puzzle. Are you sure you didn't intend "
                                     "to use read?")
@@ -183,8 +210,12 @@ class Puzzle:
         self.solution_state = puzzle_data[12]
 
         self.version = self.fileversion[:3]
-        self.solution = s.read(self.width * self.height).decode(ENCODING)
-        self.fill = s.read(self.width * self.height).decode(ENCODING)
+        # Once we have fileversion we can guess the encoding
+        self.encoding = ENCODING if self.version_tuple()[0] < 2 else ENCODING_UTF8
+        s.encoding = self.encoding
+
+        self.solution = s.read(self.width * self.height).decode(self.encoding)
+        self.fill = s.read(self.width * self.height).decode(self.encoding)
 
         self.title = s.read_string()
         self.author = s.read_string()
@@ -222,11 +253,12 @@ class Puzzle:
                 )
 
     def save(self, filename):
+        puzzle_bytes = self.tobytes()
         with open(filename, 'wb') as f:
-            f.write(self.tobytes())
+            f.write(puzzle_bytes)
 
     def tobytes(self):
-        s = PuzzleBuffer()
+        s = PuzzleBuffer(encoding=self.encoding)
         # commit any changes from helpers
         for h in self.helpers.values():
             if 'save' in dir(h):
@@ -236,14 +268,14 @@ class Puzzle:
         s.write(self.preamble)
 
         s.pack(HEADER_FORMAT,
-               self.global_cksum(), ACROSSDOWN.encode(ENCODING),
+               self.global_cksum(), ACROSSDOWN,
                self.header_cksum(), self.magic_cksum(),
                self.fileversion, self.unk1, self.scrambled_cksum,
                self.unk2, self.width, self.height,
                len(self.clues), self.puzzletype, self.solution_state)
 
-        s.write(self.solution.encode(ENCODING))
-        s.write(self.fill.encode(ENCODING))
+        s.write(self.encode(self.solution))
+        s.write(self.encode(self.fill))
 
         s.write_string(self.title)
         s.write_string(self.author)
@@ -270,34 +302,63 @@ class Puzzle:
             s.pack(EXTENSION_HEADER_FORMAT, code, len(data), data_cksum(data))
             s.write(data + b'\0')
 
-        s.write(self.postscript.encode(ENCODING))
+        # postscript is initialized, read, and stored as bytes. In case it is
+        # overwritten as a string, this try/except converts it back.
+        try:
+            postscript_bytes = self.encode(self.postscript)
+        except AttributeError:
+            postscript_bytes = self.postscript
+
+        s.write(postscript_bytes)
 
         return s.tobytes()
 
+    def encode(self, s):
+        return s.encode(self.encoding, ENCODING_ERRORS)
+
+    def encode_zstring(self, s):
+        return self.encode(s) + b'\0'
+
+    def version_tuple(self):
+        return tuple(map(int, self.version.split(b'.')))
+
+    def set_version(self, version):
+        if not isinstance(version, bytes):
+            version = bytes(version.encode('utf-8'))
+        self.version = version
+        self.fileversion = version + b'\0'
+
     def has_rebus(self):
         return self.rebus().has_rebus()
 
     def rebus(self):
-        return self.helpers.setdefault('rebus', Rebus(self))
+        if 'rebus' not in self.helpers:
+            self.helpers['rebus'] = Rebus(self)
+        return self.helpers['rebus']
 
     def has_markup(self):
         return self.markup().has_markup()
 
     def markup(self):
-        return self.helpers.setdefault('markup', Markup(self))
+        if 'markup' not in self.helpers:
+            self.helpers['markup'] = Markup(self)
+        return self.helpers['markup']
 
     def clue_numbering(self):
-        numbering = DefaultClueNumbering(self.fill, self.clues,
-                                         self.width, self.height)
-        return self.helpers.setdefault('clues', numbering)
+        if 'clues' not in self.helpers:
+            self.helpers['clues'] = DefaultClueNumbering(self.fill, self.clues, self.width, self.height)
+        return self.helpers['clues']
+
+    def blacksquare(self):
+        return BLACKSQUARE2 if self.puzzletype == PuzzleType.Diagramless else BLACKSQUARE
 
     def is_solution_locked(self):
-        return bool(self.solution_state != SolutionState.Unlocked)
+        return bool(self.solution_state == SolutionState.Locked)
 
     def unlock_solution(self, key):
         if self.is_solution_locked():
-            unscrambled = unscramble_solution(self.solution,
-                                              self.width, self.height, key)
+            unscrambled = unscramble_solution(self.solution, self.width, self.height, key,
+                                              ignore_chars=self.blacksquare())
             if not self.check_answers(unscrambled):
                 return False
 
@@ -311,16 +372,17 @@ class Puzzle:
     def lock_solution(self, key):
         if not self.is_solution_locked():
             # set the scrambled bit and cksum
-            self.scrambled_cksum = scrambled_cksum(self.solution,
-                                                   self.width, self.height)
+            self.scrambled_cksum = scrambled_cksum(self.solution, self.width, self.height,
+                                                   ignore_chars=self.blacksquare(), encoding=self.encoding)
             self.solution_state = SolutionState.Locked
-            scrambled = scramble_solution(self.solution,
-                                          self.width, self.height, key)
+            scrambled = scramble_solution(self.solution, self.width, self.height, key,
+                                          ignore_chars=self.blacksquare())
             self.solution = scrambled
 
     def check_answers(self, fill):
         if self.is_solution_locked():
-            scrambled = scrambled_cksum(fill, self.width, self.height)
+            scrambled = scrambled_cksum(fill, self.width, self.height,
+                                        ignore_chars=self.blacksquare(), encoding=self.encoding)
             return scrambled == self.scrambled_cksum
         else:
             return fill == self.solution
@@ -333,28 +395,28 @@ class Puzzle:
     def text_cksum(self, cksum=0):
         # for the checksum to work these fields must be added in order with
         # null termination, followed by all non-empty clues without null
-        # termination, followed by notes (but only for version 1.3)
+        # termination, followed by notes (but only for version >= 1.3)
         if self.title:
-            cksum = data_cksum(self.title.encode(ENCODING) + b'\0', cksum)
+            cksum = data_cksum(self.encode_zstring(self.title), cksum)
         if self.author:
-            cksum = data_cksum(self.author.encode(ENCODING) + b'\0', cksum)
+            cksum = data_cksum(self.encode_zstring(self.author), cksum)
         if self.copyright:
-            cksum = data_cksum(self.copyright.encode(ENCODING) + b'\0', cksum)
+            cksum = data_cksum(self.encode_zstring(self.copyright), cksum)
 
         for clue in self.clues:
             if clue:
-                cksum = data_cksum(clue.encode(ENCODING), cksum)
+                cksum = data_cksum(self.encode(clue), cksum)
 
-        # notes included in global cksum only in v1.3 of format
-        if self.version.decode(ENCODING) == '1.3' and self.notes:
-            cksum = data_cksum(self.notes.encode(ENCODING) + b'\0', cksum)
+        # notes included in global cksum starting v1.3 of format
+        if self.version_tuple() >= (1, 3) and self.notes:
+            cksum = data_cksum(self.encode_zstring(self.notes), cksum)
 
         return cksum
 
     def global_cksum(self):
         cksum = self.header_cksum()
-        cksum = data_cksum(self.solution.encode(ENCODING), cksum)
-        cksum = data_cksum(self.fill.encode(ENCODING), cksum)
+        cksum = data_cksum(self.encode(self.solution), cksum)
+        cksum = data_cksum(self.encode(self.fill), cksum)
         cksum = self.text_cksum(cksum)
         # extensions do not seem to be included in global cksum
         return cksum
@@ -362,8 +424,8 @@ class Puzzle:
     def magic_cksum(self):
         cksums = [
             self.header_cksum(),
-            data_cksum(self.solution.encode(ENCODING)),
-            data_cksum(self.fill.encode(ENCODING)),
+            data_cksum(self.encode(self.solution)),
+            data_cksum(self.encode(self.fill)),
             self.text_cksum()
         ]
 
@@ -385,8 +447,9 @@ class PuzzleBuffer:
     wraps a data buffer ('' or []) and provides .puz-specific methods for
     reading and writing data
     """
-    def __init__(self, data=None):
+    def __init__(self, data=None, encoding=ENCODING):
         self.data = data or []
+        self.encoding = encoding
         self.pos = 0
 
     def can_read(self, n_bytes=1):
@@ -411,7 +474,7 @@ class PuzzleBuffer:
     def read_until(self, c):
         start = self.pos
         self.seek_to(c, 1)  # read past
-        return str(self.data[start:self.pos-1], ENCODING)
+        return str(self.data[start:self.pos-1], self.encoding)
 
     def seek_to(self, s, offset=0):
         try:
@@ -427,7 +490,7 @@ class PuzzleBuffer:
 
     def write_string(self, s):
         s = s or ''
-        self.data.append(s.encode(ENCODING) + b'\0')
+        self.data.append(s.encode(self.encoding, ENCODING_ERRORS) + b'\0')
 
     def pack(self, struct_format, *values):
         self.data.append(struct.pack(struct_format, *values))
@@ -452,6 +515,63 @@ class PuzzleBuffer:
 
 
 # clue numbering helper
+def get_grid_numbering(grid, width, height):
+    # Add numbers to the grid based on positions of black squares
+    def col(index):
+        return index % width
+
+    def row(index):
+        return int(math.floor(index / width))
+
+    def len_across(index):
+        for c in range(0, width - col(index)):
+            if is_blacksquare(grid[index + c]):
+                return c
+        return c + 1
+
+    def len_down(index):
+        for c in range(0, height - row(index)):
+            if is_blacksquare(grid[index + c*width]):
+                return c
+        return c + 1
+
+    a = []
+    d = []
+    c = 0
+    n = 1
+    for i in range(0, len(grid)):
+        if not is_blacksquare(grid[i]):
+            lastc = c
+            is_across = col(i) == 0 or is_blacksquare(grid[i - 1])
+            if is_across and len_across(i) > 1:
+                a.append({
+                    'num': n,
+                    'clue': None,  # filled in by caller
+                    'clue_index': c,
+                    'cell': i,
+                    'row': row(i),
+                    'col': col(i),
+                    'len': len_across(i),
+                    'dir': 'across',
+                })
+                c += 1
+            is_down = row(i) == 0 or is_blacksquare(grid[i - width])
+            if is_down and len_down(i) > 1:
+                d.append({
+                    'num': n,
+                    'clue': None,  # filled in by caller
+                    'clue_index': c,
+                    'cell': i,
+                    'row': row(i),
+                    'col': col(i),
+                    'len': len_down(i),
+                    'dir': 'down'
+                })
+                c += 1
+            if c > lastc:
+                n += 1
+    return a, d
+
 
 class DefaultClueNumbering:
     def __init__(self, grid, clues, width, height):
@@ -460,38 +580,14 @@ class DefaultClueNumbering:
         self.width = width
         self.height = height
 
-        # compute across & down
-        a = []
-        d = []
-        c = 0
-        n = 1
-        for i in range(0, len(grid)):
-            if not is_blacksquare(grid[i]):
-                lastc = c
-                is_across = self.col(i) == 0 or is_blacksquare(grid[i - 1])
-                if is_across and self.len_across(i) > 1:
-                    a.append({
-                        'num': n,
-                        'clue': clues[c],
-                        'cell': i,
-                        'len': self.len_across(i)
-                    })
-                    c += 1
-                is_down = self.row(i) == 0 or is_blacksquare(grid[i - width])
-                if is_down and self.len_down(i) > 1:
-                    d.append({
-                        'num': n,
-                        'clue': clues[c],
-                        'cell': i,
-                        'len': self.len_down(i)
-                    })
-                    c += 1
-                if c > lastc:
-                    n += 1
-
-        self.across = a
-        self.down = d
+        self.across, self.down = get_grid_numbering(grid, width, height)
+        for entry in self.across:
+            entry['clue'] = clues[entry['clue_index']]
+        for entry in self.down:
+            entry['clue'] = clues[entry['clue_index']]
 
+    # The following methods are no longer in use, but left here in case
+    # anyone was using them externally. They may be removed in a future release.
     def col(self, index):
         return index % self.width
 
@@ -511,27 +607,85 @@ class DefaultClueNumbering:
         return c + 1
 
 
+class Grid:
+    def __init__(self, grid, width, height):
+        self.grid = grid
+        self.width = width
+        self.height = height
+        assert len(self.grid) == self.width * self.height
+
+    def get_cell(self, row, col):
+        return self.grid[self.get_cell_index(row, col)]
+
+    def get_cell_index(self, row, col):
+        return row * self.width + col
+
+    def get_range(self, row, col, length, dir='across'):
+        if dir == 'across':
+            return self.get_range_across(row, col, length)
+        elif dir == 'down':
+            return self.get_range_down(row, col, length)
+        else:
+            assert False, "dir not one of 'across' or 'down'"
+
+    def get_range_across(self, row, col, length):
+        return [self.grid[self.get_cell_index(row, col + i)] for i in range(length)]
+
+    def get_range_down(self, row, col, length):
+        return [self.grid[self.get_cell_index(row + i, col)] for i in range(length)]
+
+    def get_range_for_clue(self, clue):
+        return self.get_range(clue['row'], clue['col'], clue['len'], clue['dir'])
+
+    def get_row(self, row):
+        return self.get_range_across(row, 0, self.width)
+
+    def get_column(self, col):
+        return self.get_range_down(0, col, self.height)
+
+    def get_string(self, row, col, length, dir='across'):
+        return ''.join(self.get_range(row, col, length, dir))
+
+    def get_string_across(self, row, col, length):
+        return ''.join(self.get_range_across(row, col, length))
+
+    def get_string_down(self, row, col, length):
+        return ''.join(self.get_range_down(row, col, length))
+
+    def get_string_for_clue(self, clue):
+        return ''.join(self.get_range_for_clue(clue))
+
+
 class Rebus:
     def __init__(self, puzzle):
         self.puzzle = puzzle
+        self.fill = {}
+        self.solutions = {}
+        self.table = [0] * (self.puzzle.width * self.puzzle.height)
+
         # parse rebus data
-        rebus_data = self.puzzle.extensions.get(Extensions.Rebus, b'')
-        self.table = parse_bytes(rebus_data)
-        r_sol_data = self.puzzle.extensions.get(Extensions.RebusSolutions, b'')
-        solutions_str = r_sol_data.decode(ENCODING)
-        fill_data = self.puzzle.extensions.get(Extensions.RebusFill, b'')
-        fill_str = fill_data.decode(ENCODING)
-        self.solutions = {
-            int(item[0]): item[1]
-            for item in parse_dict(solutions_str).items()
-        }
-        self.fill = {
-            int(item[0]): item[1]
-            for item in parse_dict(fill_str).items()
-        }
+        if Extensions.Rebus in self.puzzle.extensions:
+            rebus_data = self.puzzle.extensions[Extensions.Rebus]
+            self.table = parse_bytes(rebus_data)
+
+        if Extensions.RebusSolutions in self.puzzle.extensions:
+            raw_solution_data = self.puzzle.extensions[Extensions.RebusSolutions]
+            solutions_str = raw_solution_data.decode(puzzle.encoding)
+            self.solutions = {
+                int(item[0]): item[1]
+                for item in parse_dict(solutions_str).items()
+            }
+
+        if Extensions.RebusFill in self.puzzle.extensions:
+            raw_fill_data = self.puzzle.extensions[Extensions.RebusFill]
+            fill_str = raw_fill_data.decode(puzzle.encoding)
+            self.fill = {
+                int(item[0]): item[1]
+                for item in parse_dict(fill_str).items()
+            }
 
     def has_rebus(self):
-        return Extensions.Rebus in self.puzzle.extensions
+        return Extensions.Rebus in self.puzzle.extensions or any(self.table)
 
     def is_rebus_square(self, index):
         return bool(self.table[index])
@@ -557,9 +711,9 @@ class Rebus:
         if self.has_rebus():
             # commit changes back to puzzle.extensions
             self.puzzle.extensions[Extensions.Rebus] = pack_bytes(self.table)
-            rebus_solutions = dict_to_string(self.solutions).encode(ENCODING)
+            rebus_solutions = self.puzzle.encode(dict_to_string(self.solutions))
             self.puzzle.extensions[Extensions.RebusSolutions] = rebus_solutions
-            rebus_fill = dict_to_string(self.fill).encode(ENCODING)
+            rebus_fill = self.puzzle.encode(dict_to_string(self.fill))
             self.puzzle.extensions[Extensions.RebusFill] = rebus_fill
 
 
@@ -577,7 +731,7 @@ class Markup:
         return [i for i, b in enumerate(self.markup) if b]
 
     def is_markup_square(self, index):
-        return bool(self.table[index])
+        return bool(self.markup[index])
 
     def save(self):
         if self.has_markup():
@@ -601,9 +755,15 @@ def data_cksum(data, cksum=0):
     return cksum
 
 
-def scramble_solution(solution, width, height, key):
+def replace_chars(s, chars, replacement=''):
+    for ch in chars:
+        s = s.replace(ch, replacement)
+    return s
+
+
+def scramble_solution(solution, width, height, key, ignore_chars=BLACKSQUARE):
     sq = square(solution, width, height)
-    data = restore(sq, scramble_string(sq.replace(BLACKSQUARE, ''), key))
+    data = restore(sq, scramble_string(replace_chars(sq, ignore_chars), key))
     return square(data, height, width)
 
 
@@ -629,16 +789,16 @@ def scramble_string(s, key):
     return s
 
 
-def unscramble_solution(scrambled, width, height, key):
+def unscramble_solution(scrambled, width, height, key, ignore_chars=BLACKSQUARE):
     # width and height are reversed here
     sq = square(scrambled, width, height)
-    data = restore(sq, unscramble_string(sq.replace(BLACKSQUARE, ''), key))
+    data = restore(sq, unscramble_string(replace_chars(sq, ignore_chars), key))
     return square(data, height, width)
 
 
 def unscramble_string(s, key):
     key = key_digits(key)
-    l = len(s)
+    l = len(s)  # noqa: E741
     for k in key[::-1]:
         s = unshuffle(s)
         s = s[l-k:] + s[:l-k]
@@ -647,9 +807,9 @@ def unscramble_string(s, key):
     return s
 
 
-def scrambled_cksum(scrambled, width, height):
-    data = square(scrambled, width, height).replace(BLACKSQUARE, '')
-    return data_cksum(data.encode(ENCODING))
+def scrambled_cksum(scrambled, width, height, ignore_chars=BLACKSQUARE, encoding=ENCODING):
+    data = replace_chars(square(scrambled, width, height), ignore_chars)
+    return data_cksum(data.encode(encoding, ENCODING_ERRORS))
 
 
 def key_digits(key):
@@ -703,7 +863,7 @@ def restore(s, t):
 def is_blacksquare(c):
     if isinstance(c, int):
         c = chr(c)
-    return c == BLACKSQUARE
+    return c in [BLACKSQUARE, BLACKSQUARE2]
 
 
 #
@@ -722,8 +882,120 @@ def pack_bytes(a):
 # dict string format is k1:v1;k2:v2;...;kn:vn;
 # (for whatever reason there's a trailing ';')
 def parse_dict(s):
-    return dict(p.split(':') for p in s.split(';') if ':' in p)
+    return dict(p.split(':', 1) for p in s.split(';') if ':' in p)
 
 
 def dict_to_string(d):
     return ';'.join(':'.join(map(str, [k, v])) for k, v in d.items()) + ';'
+
+
+def from_text_format(s):
+    d = text_file_as_dict(s)
+
+    if 'ACROSS PUZZLE' in d:
+        # file_version = 'v1'
+        pass
+    elif 'ACROSS PUZZLE v2' in d:
+        # file_version = 'v2'
+        pass
+    else:
+        raise PuzzleFormatError('Not a valid Across Lite text puzzle')
+
+    p = Puzzle()
+    across_clues = []
+    down_clues = []
+    if 'TITLE' in d:
+        p.title = d['TITLE']
+    if 'AUTHOR' in d:
+        p.author = d['AUTHOR']
+    if 'COPYRIGHT' in d:
+        p.copyright = d['COPYRIGHT']
+    if 'SIZE' in d:
+        size = d['SIZE'].split('x')
+        p.width = int(size[0])
+        p.height = int(size[1])
+    if 'GRID' in d:
+        solution_lines = d['GRID'].splitlines()
+        p.solution = ''.join(line.strip() for line in solution_lines if line.strip())
+    if 'ACROSS' in d:
+        across_clues.extend(line.strip() for line in d['ACROSS'].splitlines() if line.strip())
+    if 'DOWN' in d:
+        down_clues.extend(line.strip() for line in d['DOWN'].splitlines() if line.strip())
+    if 'NOTEPAD' in d:
+        p.notes = d['NOTEPAD']
+    if 'REBUS' in d:
+        pass  # TODO: text file REBUS
+
+    if p.solution:
+        p.fill = ''.join(c if c == BLACKSQUARE else BLANKSQUARE for c in p.solution)
+        across, down = get_grid_numbering(p.fill, p.width, p.height)
+        # we have to match puzfile's expected clue ordering or we won't be able to
+        # write the puzzle out as a valid .puz file
+        p.clues = [''] * (len(across) + len(down))
+        for i in range(len(across)):
+            clue = across_clues[i] if i < len(across_clues) else ''
+            across[i]['clue'] = clue
+            p.clues[across[i]['clue_index']] = clue
+        for i in range(len(down)):
+            clue = down_clues[i] if i < len(down_clues) else ''
+            down[i]['clue'] = clue
+            p.clues[down[i]['clue_index']] = clue
+
+    return p
+
+
+def text_file_as_dict(s):
+    d = {}
+    k = None
+    v = []
+    for line in s.splitlines():
+        line = line.strip()
+        if line.startswith('<') and line.endswith('>'):
+            if k:
+                d[k] = '\n'.join(v)
+            k = line[1:-1]
+            v = []
+        else:
+            v.append(line)
+
+    if k:
+        d[k] = '\n'.join(v)
+    return d
+
+
+def to_text_format(p, text_version='v1'):
+    TAB = '\t'  # most lines begin indented with whitespace
+    lines = []
+    if text_version == 'v1':
+        lines.append('<ACROSS PUZZLE>')
+    elif text_version:
+        lines.append(f'<ACROSS PUZZLE {text_version}>')
+    else:
+        raise ValueError("invalid text_version")
+
+    lines.append('<TITLE>')
+    lines.append(TAB + p.title)
+    lines.append('<AUTHOR>')
+    lines.append(TAB + p.author)
+    lines.append('<COPYRIGHT>')
+    lines.append(TAB + p.copyright)
+    lines.append('<SIZE>')
+    lines.append(TAB + f'{p.width}x{p.height}')
+    lines.append('<GRID>')
+    for r in range(p.height):
+        row = p.solution[r*p.width:(r+1)*p.width]
+        lines.append(TAB + row)
+
+    # get clues in across/down order
+    numbering = p.clue_numbering()
+    lines.append('<ACROSS>')
+    for clue in numbering.across:
+        lines.append(TAB + clue['clue'])
+    lines.append('<DOWN>')
+    for clue in numbering.down:
+        lines.append(TAB + clue['clue'])
+
+    lines.append('<NOTEPAD>')
+    lines.append(p.notes)  # no tab here, idk why
+
+    return '\n'.join(lines)
